--============================================================================--
--! @file       rmq_stream_fifo.vhd
--! @author     Alen Arias Vazquez <alen.arias.vazquez@cern.ch>
--! @company    CERN
--! @date       2023-03-01
--! @brief      Fifo encapsulation of RMQ frames
--============================================================================--


--============================================================================--
--                            Libraries & Packages                            --
--============================================================================--
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.mt_mqueue_pkg.all;
use work.gencores_pkg.all;

--============================================================================--
--               Entity declaration for rmq_stream_fifo                       --
--============================================================================--
entity rmq_stream_fifo is
    generic (
        g_DATA_WIDTH        : natural;
        g_CDC_ENABLE        : natural range 0 to 1;
        g_FIFO_DEPTH        : natural := 128;
        g_show_ahead        : boolean := true
    );
    port (
        rst_n_i             : in  std_logic;                --! FPGA reset
        --! RX
        rmq_src_clk_i       : in  std_logic;
        rmq_src_i           : in  t_mt_stream_source_in;    --! (ready, pkt_ready)
        rmq_src_o           : out t_mt_stream_source_out;   --! (data, hdr, valid, last, error)
        rmq_src_config_i    : in  t_mt_stream_config_out;   --! (adr, dat, we)
        rmq_src_config_o    : out t_mt_stream_config_in;    --! (dat)
        --! TX
        rmq_snk_clk_i       : in  std_logic;
        rmq_snk_i           : in  t_mt_stream_sink_in;      --! (data, hdr, valid, last, error)
        rmq_snk_o           : out t_mt_stream_sink_out;     --! (ready, pkt_ready)
        rmq_snk_config_i    : in  t_mt_stream_config_out;   --! (adr, dat, we)
        rmq_snk_config_o    : out t_mt_stream_config_in     --! (dat)
    );
end entity rmq_stream_fifo;

--============================================================================--
--                           Architecture declaration                         --
--============================================================================--
architecture struct of rmq_stream_fifo is

    --! Data encapsulation
    signal s_fifo_input     : std_logic_vector(g_DATA_WIDTH+2 downto 0);
    signal s_fifo_output    : std_logic_vector(g_DATA_WIDTH+2 downto 0);

    --! Write enable
    signal s_input_full     : std_logic;
    signal s_write_enable   : std_logic;

    --! Read enable
    signal s_output_empty   : std_logic;
    signal s_read_enable    : std_logic;

--------------------------------------------------------------------------------
-- Architecture Begin
--------------------------------------------------------------------------------
begin

    ----------------------------------------------------------------------------
    --! Unused
    rmq_src_config_o.dat <= (others => '0');
    rmq_snk_config_o.dat <= (others => '0');

    ----------------------------------------------------------------------------
    CDC_check: assert g_CDC_ENABLE = 0
        report "The port rmq_snk_clk_i is ignored" severity note;

    s_fifo_input <= rmq_snk_i.hdr & rmq_snk_i.last & rmq_snk_i.error & rmq_snk_i.data;
    s_write_enable <= not(s_input_full) and rmq_snk_i.valid;

    rmq_snk_o.ready <= not (s_input_full);
    rmq_snk_o.pkt_ready <= not (s_input_full);

    gen_CDC_false: if g_CDC_ENABLE = 0 generate
        cmp_fifo: entity work.inferred_sync_fifo
        generic map (
            g_data_width                => g_DATA_WIDTH+3,
            g_size                      => g_FIFO_DEPTH,
            g_show_ahead                => g_show_ahead,
            g_show_ahead_legacy_mode    => true,
            g_with_empty                => true,
            g_with_full                 => true,
            g_with_almost_empty         => false,
            g_with_almost_full          => false,
            g_with_count                => false,
            g_register_flag_outputs     => false,
            g_almost_empty_threshold    => 0,
            g_almost_full_threshold     => 0
        )
        port map(
            rst_n_i                     => rst_n_i,
            clk_i                       => rmq_src_clk_i,
            d_i                         => s_fifo_input,
            we_i                        => s_write_enable,
            full_o                      => s_input_full,
            q_o                         => s_fifo_output,
            empty_o                     => s_output_empty,
            rd_i                        => s_read_enable,
            almost_empty_o              => open,
            almost_full_o               => open,
            count_o                     => open
        );
    end generate gen_CDC_false;

    gen_CDC_true: if g_CDC_ENABLE = 1 generate
        cmp_fifo: entity work.inferred_async_fifo
        generic map (
            g_data_width                => g_DATA_WIDTH+3,
            g_size                      => g_FIFO_DEPTH,
            g_show_ahead                => true,
            g_with_rd_empty             => true,
            g_with_rd_full              => false,
            g_with_rd_almost_empty      => false,
            g_with_rd_almost_full       => false,
            g_with_rd_count             => false,
            g_with_wr_full              => true,
            g_with_wr_empty             => false,
            g_with_wr_almost_empty      => false,
            g_with_wr_almost_full       => false,
            g_with_wr_count             => false,
            g_almost_empty_threshold    => 0,
            g_almost_full_threshold     => 0
        )
        port map (
            rst_n_i                     => rst_n_i,
            -- write port
            clk_wr_i                    => rmq_snk_clk_i,
            d_i                         => s_fifo_input,
            we_i                        => s_write_enable,
            wr_full_o                   => s_input_full,
            wr_empty_o                  => open,
            wr_almost_empty_o           => open,
            wr_almost_full_o            => open,
            wr_count_o                  => open,
            -- read port
            clk_rd_i                    => rmq_src_clk_i,
            q_o                         => s_fifo_output,
            rd_empty_o                  => s_output_empty,
            rd_i                        => s_read_enable,
            rd_full_o                   => open,
            rd_almost_empty_o           => open,
            rd_almost_full_o            => open,
            rd_count_o                  => open
        );
    end generate gen_CDC_true;

    rmq_src_o.data  <= s_fifo_output(g_DATA_WIDTH-1 downto 0);
    rmq_src_o.hdr   <= s_fifo_output(g_DATA_WIDTH+2);
    rmq_src_o.last  <= s_fifo_output(g_DATA_WIDTH+1);
    rmq_src_o.error <= s_fifo_output(g_DATA_WIDTH);
    rmq_src_o.valid <= not (s_output_empty);

    s_read_enable <= rmq_src_i.ready and not(s_output_empty);

end architecture struct;
--============================================================================--
--                           End Architecture                                 --
--============================================================================--
