--============================================================================--
--! @file       spi_rmq_bridge.vhd
--! @author     Alen Arias Vazquez <alen.arias.vazquez@cern.ch>
--! @company    CERN
--! @date       2023-02-21
--! @brief      Bridge between SPI and RMQ
--============================================================================--


--============================================================================--
--                            Libraries & Packages                            --
--============================================================================--
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.mock_turtle_pkg.all;
use work.gencores_pkg.all;
use work.wishbone_pkg.all;
use work.mt_mqueue_pkg.all;

--============================================================================--
--               Entity declaration for spi_rmq_bridge                        --
--============================================================================--
entity spi_rmq_bridge is
    generic (
        g_DATA_WIDTH        : natural;
        g_CPOL              : natural range 0 to 1;
        g_CPHA              : natural range 0 to 1;
        g_CDC_ENABLE        : natural range 0 to 1;
        g_INPUT_FIFO_DEPTH  : natural := 512;
        g_OUTPUT_FIFO_DEPTH : natural := 1024
    );
    port (
        rst_n_i             : in  std_logic;                --! FPGA reset
        clk_i               : in  std_logic;                --! FPGA clock
        spi_sample_clk_i    : in  std_logic;                --! FPGA clock faster
        --! RX (SPI -> RMQ)
        rmq_src_i           : in  t_mt_stream_source_in;    --! (ready, pkt_ready)
        rmq_src_o           : out t_mt_stream_source_out;   --! (data, hdr, valid, last, error)
        rmq_src_config_i    : in  t_mt_stream_config_out;   --! (adr, dat, we)
        rmq_src_config_o    : out t_mt_stream_config_in;    --! (dat)
        --! TX (RMQ -> SPI)
        rmq_snk_i           : in  t_mt_stream_sink_in;      --! (data, hdr, valid, last, error)
        rmq_snk_o           : out t_mt_stream_sink_out;     --! (ready, pkt_ready)
        rmq_snk_config_i    : in  t_mt_stream_config_out;   --! (adr, dat, we)
        rmq_snk_config_o    : out t_mt_stream_config_in;    --! (dat)
        --! SPI interface
        spi_clk_i           : in  std_logic;
        spi_cs_n_i          : in  std_logic;
        spi_mosi_i          : in  std_logic;
        spi_miso_o          : out std_logic
    );
end entity spi_rmq_bridge;

--============================================================================--
--                           Architecture declaration                         --
--============================================================================--
architecture rtl of spi_rmq_bridge is

    --! SPI Synch:
    signal s_spi_cs_n           : std_logic;
    signal s_spi_mosi           : std_logic;
    signal s_spi_clk_re         : std_logic;
    signal s_spi_clk_fe         : std_logic;

    attribute fsm_encoding      : string;

    --! SPI FSM:
    type t_SPI_STATE is (RX_SPI_HEADER, PROCESSING, WR_RMQ, RD_RMQ, CS_DEASSERT);
    signal s_spi_state                          : t_SPI_STATE;
    attribute fsm_encoding of s_spi_state       : signal is "one_hot";

    --! WR RMQ:
    type t_WR_RMQ is (WR_HEADER, WR_MSB, WR_LSB);
    signal s_wr_rmq_state                       : t_WR_RMQ;
    attribute fsm_encoding of s_wr_rmq_state    : signal is "one_hot";

    --! RD RMQ:
    type t_RD_RMQ is (RD_HEADER, RD_DATA_LSB, RD_DATA_MSB);
    signal s_rd_rmq_state                       : t_RD_RMQ;
    attribute fsm_encoding of s_rd_rmq_state    : signal is "one_hot";

    --! Signals FSM
    signal s_tx_shreg           : std_logic_vector(g_DATA_WIDTH-1 downto 0);
    signal s_rx_shreg           : std_logic_vector(g_DATA_WIDTH-1 downto 0);
    signal s_tx_bit_c           : unsigned(f_log2_ceil(g_DATA_WIDTH)-1 downto 0);
    signal s_rx_bit_c           : unsigned(f_log2_ceil(g_DATA_WIDTH)-1 downto 0);
    signal s_rmq_id             : std_logic_vector(7 downto 0);
    signal s_data_len           : std_logic_vector(7 downto 0);
    signal s_crtl_flags         : std_logic_vector(7 downto 0);
    signal s_wr_word_c          : unsigned(8 downto 0);
    signal s_operation          : std_logic; -- 0 write 1 read
    signal s_ready              : std_logic;
    signal s_read_last          : std_logic;

    signal s_data_header        : std_logic;
    signal s_data_valid         : std_logic;
    signal s_data_last          : std_logic;
    signal s_data_error         : std_logic;

    signal sample_en            : std_logic;
    signal shift_en             : std_logic;

    --! FIFO MOSI
    signal s_mosi_lsb           : std_logic_vector((g_data_width/2)-1 downto 0);
    signal s_mosi_msb           : std_logic_vector((g_data_width/2)-1 downto 0);
    signal s_mosi_header        : std_logic;
    signal s_mosi_valid         : std_logic;
    signal s_mosi_last          : std_logic;
    signal s_mosi_error         : std_logic;
    signal s_mosi_fifo_data_in  : std_logic_vector(g_data_width+2 downto 0);
    signal s_mosi_fifo_data_out : std_logic_vector(g_data_width+2 downto 0);
    signal s_mosi_wr_en         : std_logic;
    signal s_mosi_rd_en         : std_logic;
    signal s_mosi_fifo_wr_full  : std_logic;
    signal s_mosi_fifo_rd_empty : std_logic;

    --! FIFO MISO
    signal s_miso_fifo_data_in  : std_logic_vector(g_data_width downto 0);
    signal s_miso_fifo_data_out : std_logic_vector(g_data_width downto 0);
    signal s_miso_wr_en         : std_logic;
    signal s_miso_rd_en         : std_logic;
    signal s_miso_fifo_wr_full  : std_logic;
    signal s_miso_fifo_rd_empty : std_logic;

--------------------------------------------------------------------------------
-- Architecture Begin
--------------------------------------------------------------------------------
begin

    ----------------------------------------------------------------------------
    --! Synch CLK
    cmp_clk_sync_ffs : entity work.gc_sync_ffs
        generic map (
            g_SYNC_EDGE => "positive"
        )
        port map (
            clk_i    => spi_sample_clk_i,
            rst_n_i  => rst_n_i,
            data_i   => spi_clk_i,
            synced_o => open,
            npulse_o => s_spi_clk_fe,
            ppulse_o => s_spi_clk_re
        );

    --! Synch CSn
    cmp_cs_n_sync_ffs : entity work.gc_sync_ffs
        generic map (
            g_SYNC_EDGE => "positive"
        )
        port map (
            clk_i    => spi_sample_clk_i,
            rst_n_i  => rst_n_i,
            data_i   => spi_cs_n_i,
            synced_o => s_spi_cs_n,
            npulse_o => open,
            ppulse_o => open
        );

    --! Synch MOSI
    cmp_mosi_sync_ffs : entity work.gc_sync_ffs
        generic map (
            g_SYNC_EDGE => "positive"
        )
        port map (
            clk_i    => spi_sample_clk_i,
            rst_n_i  => rst_n_i,
            data_i   => spi_mosi_i,
            synced_o => s_spi_mosi,
            npulse_o => open,
            ppulse_o => open
        );

    -- Define the sampling and shifting based on the CPHA
    gen_pha_zero : if g_CPHA = 0 generate
        sample_en <= s_spi_clk_re;
        shift_en  <= s_spi_clk_fe;
    end generate;

    gen_pha_one : if g_CPHA = 1 generate
        sample_en <= s_spi_clk_fe;
        shift_en  <= s_spi_clk_re;
    end generate;

    ----------------------------------------------------------------------------
    --! SPI FSM
    p_spi_fsm : process(spi_sample_clk_i)
    begin
        if rising_edge(spi_sample_clk_i) then
            if rst_n_i = '0' then

                s_spi_state <= RX_SPI_HEADER;
                s_tx_shreg <= (others => '0');
                s_rx_shreg <= (others => '0');
                s_tx_bit_c <= (others => '0');
                s_rx_bit_c <= (others => '0');
                s_rmq_id <= (others => '0');
                s_data_len <= (others => '0');
                s_crtl_flags <= (others => '0');
                s_wr_word_c <= (others => '0');
                s_data_valid <= '0';
                s_data_header <= '0';
                s_data_last <= '0';
                s_data_error <= '0';
                s_ready <= '0';
                s_read_last <= '0';
                s_operation <= '0';

            else

                case s_spi_state is

                    ------------------------------------------
                    when RX_SPI_HEADER =>
                        s_ready <= '0';
                        s_read_last <= '0';
                        s_data_valid <= '0';
                        if s_spi_cs_n = '0' then
                            if sample_en = '1' then
                                if s_rx_bit_c = 7 then
                                    s_operation <= s_spi_mosi;
                                end if;
                                s_rx_shreg <= s_rx_shreg(g_data_width-2 downto 0) & s_spi_mosi;
                                if (s_rx_bit_c = g_data_width-1) then
                                    s_rx_bit_c <= (others=>'0');
                                    s_spi_state <= PROCESSING;
                                else
                                    s_rx_bit_c <= s_rx_bit_c + 1;
                                    s_spi_state <= RX_SPI_HEADER;
                                end if;
                            end if;
                        else
                            s_spi_state <= RX_SPI_HEADER;
                            s_tx_shreg <= (others => '0');
                            s_rx_shreg <= (others => '0');
                            s_tx_bit_c <= (others => '0');
                            s_rx_bit_c <= (others => '0');
                        end if;

                    ------------------------------------------
                    when PROCESSING =>
                        if s_spi_cs_n = '0' then
                            s_data_len <= s_rx_shreg(15 downto 8);
                            s_rmq_id <= s_rx_shreg(23 downto 16);
                            if s_operation = '1' then
                                s_spi_state <= RD_RMQ;
                            else
                                s_spi_state <= WR_RMQ;
                            end if;
                        else
                            s_spi_state <= WR_RMQ;
                        end if;

                    ------------------------------------------
                    when WR_RMQ =>
                        if s_spi_cs_n = '0' then
                            s_data_error <= '0';
                            s_data_valid <= '0';
                            s_data_header <= '0';
                            s_data_last <= '0';
                            if sample_en = '1' then
                                s_rx_shreg <= s_rx_shreg(g_data_width-2 downto 0) & s_spi_mosi;
                                if (s_rx_bit_c = g_data_width-1) then
                                    s_rx_bit_c <= (others=>'0');
                                    s_wr_word_c <= s_wr_word_c +1;
                                    s_data_valid <= '1';
                                    if s_wr_word_c = 0 then
                                        s_data_header <= '1';
                                        s_spi_state <= WR_RMQ;
                                    elsif s_wr_word_c = unsigned(s_data_len)-1 then
                                        s_data_last <= '1';
                                        s_spi_state <= RX_SPI_HEADER;
                                    else
                                        s_data_header <= '0';
                                        s_data_last <= '0';
                                        s_spi_state <= WR_RMQ;
                                    end if;
                                else
                                    s_rx_bit_c <= s_rx_bit_c + 1;
                                    s_wr_word_c <= s_wr_word_c;
                                    s_spi_state <= WR_RMQ;
                                end if;
                            end if;
                        else
                            s_data_error <= '1';
                            s_spi_state <= RX_SPI_HEADER;
                            s_tx_shreg <= (others => '0');
                            s_rx_shreg <= (others => '0');
                            s_tx_bit_c <= (others => '0');
                            s_rx_bit_c <= (others => '0');
                            s_wr_word_c <= (others => '0');
                        end if;

                    ------------------------------------------
                    when RD_RMQ =>
                        if s_spi_cs_n = '0' then
                            if shift_en = '1' then
                                if g_CPHA = 1 and s_tx_bit_c = 0 then
                                    s_tx_shreg <= s_miso_fifo_data_out(31 downto 0);
                                    s_read_last <= s_miso_fifo_data_out(32);
                                    s_tx_bit_c <= s_tx_bit_c + 1;
                                    s_ready <= '1';
                                    if s_read_last = '1' then
                                        s_ready <= '0';
                                        s_spi_state <= CS_DEASSERT;
                                    else
                                        s_ready <= '1';
                                        s_spi_state <= RD_RMQ;
                                    end if;
                                else
                                    s_tx_shreg <= s_tx_shreg(g_data_width-2 downto 0) & '0';
                                    if s_tx_bit_c = g_data_width-1 then
                                        s_tx_bit_c <= (others=>'0');
                                    else
                                        s_tx_bit_c <= s_tx_bit_c + 1;
                                    end if;
                                end if;
                            else
                                s_ready <= '0';
                            end if;
                        else
                            s_spi_state <= RX_SPI_HEADER;
                        end if;

                    ------------------------------------------
                    when CS_DEASSERT =>
                        if s_spi_cs_n = '1' then
                            s_spi_state <= RX_SPI_HEADER;
                        else
                            s_spi_state <= CS_DEASSERT;
                        end if;

                    when others =>
                        s_spi_state <= RX_SPI_HEADER;

                end case;

            end if;
        end if;
    end process p_spi_fsm;

    spi_miso_o <= s_tx_shreg(g_DATA_WIDTH-1);

    ----------------------------------------------------------------------------
    --! RD RMQ FIFO
    gen_miso_no_CDC: if g_CDC_ENABLE = 0 generate
        cmp_output_fifo:entity work.inferred_sync_fifo
        generic map (
            g_data_width                => g_DATA_WIDTH+1,
            g_size                      => g_OUTPUT_FIFO_DEPTH,
            g_show_ahead                => true,
            g_with_empty                => true,
            g_with_full                 => true,
            g_with_almost_empty         => false,
            g_with_almost_full          => false,
            g_with_count                => false,
            g_almost_empty_threshold    => 0,
            g_almost_full_threshold     => 0,
            g_register_flag_outputs     => False
        )
        port map(
            rst_n_i                     => rst_n_i,
            clk_i                       => clk_i,
            d_i                         => s_miso_fifo_data_in,
            we_i                        => s_miso_wr_en,
            full_o                      => s_miso_fifo_wr_full,
            q_o                         => s_miso_fifo_data_out,
            empty_o                     => s_miso_fifo_rd_empty,
            rd_i                        => s_miso_rd_en,
            almost_empty_o              => open,
            almost_full_o               => open,
            count_o                     => open
        );
    end generate gen_miso_no_CDC;

    gen_miso_CDC: if g_CDC_ENABLE = 1 generate
        cmp_output_fifo:entity work.inferred_async_fifo
        generic map (
            g_data_width                => g_DATA_WIDTH+1,
            g_size                      => g_OUTPUT_FIFO_DEPTH,
            g_show_ahead                => true,
            g_with_rd_empty             => true,
            g_with_rd_full              => false,
            g_with_rd_almost_empty      => false,
            g_with_rd_almost_full       => false,
            g_with_rd_count             => false,
            g_with_wr_empty             => false,
            g_with_wr_full              => true,
            g_with_wr_almost_empty      => false,
            g_with_wr_almost_full       => false,
            g_with_wr_count             => false,
            g_almost_empty_threshold    => 0,
            g_almost_full_threshold     => 0
        )
        port map (
            rst_n_i                     => rst_n_i,
            -- write port
            clk_wr_i                    => clk_i,
            d_i                         => s_miso_fifo_data_in,
            we_i                        => s_miso_wr_en,
            wr_full_o                   => s_miso_fifo_wr_full,
            wr_empty_o                  => open,
            wr_almost_empty_o           => open,
            wr_almost_full_o            => open,
            wr_count_o                  => open,
            -- read port
            clk_rd_i                    => spi_sample_clk_i,
            q_o                         => s_miso_fifo_data_out,
            rd_empty_o                  => s_miso_fifo_rd_empty,
            rd_i                        => s_miso_rd_en,
            rd_full_o                   => open,
            rd_almost_empty_o           => open,
            rd_almost_full_o            => open,
            rd_count_o                  => open
        );
    end generate gen_miso_CDC;

    s_miso_rd_en <= s_ready and not(s_miso_fifo_rd_empty);

    rmq_snk_o.ready <= not (s_miso_fifo_wr_full);
    rmq_snk_o.pkt_ready <= not (s_miso_fifo_wr_full);

    p_push_miso_fifo : process(clk_i)
    begin
        if rising_edge(clk_i) then
            if rst_n_i = '0' then
                s_miso_wr_en <= '0';
                s_miso_fifo_data_in <= (others => '0');
                s_rd_rmq_state <= RD_HEADER;
            else
                case s_rd_rmq_state is
                    ------------------------------------------
                    when RD_HEADER =>
                        if s_miso_fifo_wr_full = '0' and rmq_snk_i.valid = '1' and rmq_snk_i.hdr = '1' then
                            s_miso_fifo_data_in(31 downto 0) <= rmq_snk_i.data;
                            s_miso_fifo_data_in(32) <= '0';
                            s_miso_wr_en <= '1';
                            s_rd_rmq_state <= RD_DATA_MSB;
                        else
                            s_miso_wr_en <= '0';
                            s_rd_rmq_state <= RD_HEADER;
                        end if;

                    ------------------------------------------
                    when RD_DATA_MSB =>
                        s_miso_wr_en <= '0';
                        if rmq_snk_i.valid = '1' and rmq_snk_i.hdr = '0' then
                            s_miso_fifo_data_in(31 downto 16) <= rmq_snk_i.data(15 downto 0);
                            s_rd_rmq_state <= RD_DATA_LSB;
                        else
                            s_rd_rmq_state <= RD_DATA_MSB;
                        end if;

                    ------------------------------------------
                    when RD_DATA_LSB =>
                        if s_miso_fifo_wr_full = '0' and rmq_snk_i.valid = '1' and rmq_snk_i.hdr = '0' then
                            s_miso_fifo_data_in(15 downto 0) <= rmq_snk_i.data(15 downto 0);
                            s_miso_fifo_data_in(32) <= rmq_snk_i.last;
                            s_miso_wr_en <= '1';
                            if rmq_snk_i.last = '1' then
                                s_rd_rmq_state <= RD_HEADER;
                            else
                                s_rd_rmq_state <= RD_DATA_MSB;
                            end if;
                        else
                            s_miso_wr_en <= '0';
                            s_rd_rmq_state <= RD_DATA_LSB;
                        end if;

                    ------------------------------------------
                    when others =>
                        s_rd_rmq_state <= RD_HEADER;

                end case;
            end if;
        end if;
    end process p_push_miso_fifo;

    ----------------------------------------------------------------------------
    --! WR RMQ FIFO
    p_wr_rmq : process(spi_sample_clk_i)
    begin
        if rising_edge(spi_sample_clk_i) then
            if rst_n_i = '0' then
                s_mosi_fifo_data_in <= (others => '0');
                s_mosi_header <= '0';
                s_mosi_last <= '0';
                s_mosi_error <= '0';
                s_mosi_msb <= (others => '0');
                s_mosi_lsb <= (others => '0');
                s_mosi_wr_en <= '0';
                s_wr_rmq_state <= WR_HEADER;
            else
                case s_wr_rmq_state is
                    ------------------------------------------
                    when WR_HEADER =>
                        s_mosi_fifo_data_in <= (others => '0');
                        s_mosi_wr_en <= '0';
                        if s_data_valid = '1' then
                            s_mosi_msb <= s_rx_shreg(31 downto 16);
                            s_mosi_lsb <= s_rx_shreg(15 downto 0);
                            s_mosi_header <= s_data_header;
                            s_mosi_last <= s_data_last;
                            s_mosi_error <= s_data_error;
                            s_wr_rmq_state <= WR_MSB;
                        else
                            s_mosi_msb <= (others => '0');
                            s_mosi_lsb <= (others => '0');
                            s_mosi_header <= '0';
                            s_mosi_last <= '0';
                            s_mosi_error <= '0';
                            s_wr_rmq_state <= WR_HEADER;
                        end if;

                    ------------------------------------------
                    when WR_MSB =>
                        s_mosi_fifo_data_in <= s_mosi_header & '0' & s_mosi_error & x"0000" & s_mosi_msb;
                        s_mosi_msb <= s_mosi_msb;
                        s_mosi_lsb <= s_mosi_lsb;
                        s_mosi_last <= s_mosi_last;
                        s_mosi_error <= s_mosi_error;
                        s_mosi_header <= s_mosi_header;
                        if s_mosi_fifo_wr_full = '0' then
                            s_mosi_wr_en <= '1';
                            s_wr_rmq_state <= WR_LSB;
                        else
                            s_mosi_wr_en <= '0';
                            s_wr_rmq_state <= WR_MSB;
                        end if;

                    ------------------------------------------
                    when WR_LSB =>
                        s_mosi_fifo_data_in <= s_mosi_header & s_mosi_last & s_mosi_error & x"0000" & s_mosi_lsb;
                        if s_mosi_fifo_wr_full = '0' then
                            s_mosi_msb <= (others => '0');
                            s_mosi_lsb <= (others => '0');
                            s_mosi_header <= '0';
                            s_mosi_last <= '0';
                            s_mosi_error <= '0';
                            s_mosi_wr_en <= '1';
                            s_wr_rmq_state <= WR_HEADER;
                        else
                            s_mosi_msb <= s_mosi_msb;
                            s_mosi_lsb <= s_mosi_lsb;
                            s_mosi_header <= s_mosi_header;
                            s_mosi_last <= s_mosi_last;
                            s_mosi_error <= s_mosi_error;
                            s_mosi_wr_en <= '0';
                            s_wr_rmq_state <= WR_LSB;
                        end if;

                    ------------------------------------------
                    when others =>
                        s_mosi_fifo_data_in <= (others => '0');
                        s_mosi_msb <= (others => '0');
                        s_mosi_lsb <= (others => '0');
                        s_mosi_header <= '0';
                        s_mosi_last <= '0';
                        s_mosi_error <= '0';
                        s_mosi_wr_en <= '0';
                        s_wr_rmq_state <= WR_HEADER;

                end case;
            end if;
        end if;
    end process p_wr_rmq;

    gen_mosi_no_CDC: if g_CDC_ENABLE = 0 generate
        cmp_input_fifo:entity work.inferred_sync_fifo
        generic map (
            g_data_width                => g_DATA_WIDTH+3,
            g_size                      => g_INPUT_FIFO_DEPTH,
            g_show_ahead                => true,
            g_show_ahead_legacy_mode    => false,
            g_with_empty                => true,
            g_with_full                 => true,
            g_with_almost_empty         => false,
            g_with_almost_full          => false,
            g_with_count                => false,
            g_almost_empty_threshold    => 0,
            g_almost_full_threshold     => 0,
            g_register_flag_outputs     => False
        )
        port map(
            rst_n_i                     => rst_n_i,
            clk_i                       => spi_sample_clk_i,
            d_i                         => s_mosi_fifo_data_in,
            we_i                        => s_mosi_wr_en,
            full_o                      => s_mosi_fifo_wr_full,
            q_o                         => s_mosi_fifo_data_out,
            empty_o                     => s_mosi_fifo_rd_empty,
            rd_i                        => s_mosi_rd_en,
            almost_empty_o              => open,
            almost_full_o               => open,
            count_o                     => open
        );
    end generate gen_mosi_no_CDC;

    gen_mosi_CDC: if g_CDC_ENABLE = 1 generate
        cmp_input_fifo:entity work.inferred_async_fifo
        generic map (
            g_data_width                => g_DATA_WIDTH+3,
            g_size                      => g_INPUT_FIFO_DEPTH,
            g_show_ahead                => true,
            g_with_rd_empty             => true,
            g_with_rd_full              => false,
            g_with_rd_almost_empty      => false,
            g_with_rd_almost_full       => false,
            g_with_rd_count             => false,
            g_with_wr_empty             => false,
            g_with_wr_full              => true,
            g_with_wr_almost_empty      => false,
            g_with_wr_almost_full       => false,
            g_with_wr_count             => false,
            g_almost_empty_threshold    => 0,
            g_almost_full_threshold     => 0
        )
        port map (
            rst_n_i                     => rst_n_i,
            -- write port
            clk_wr_i                    => spi_sample_clk_i,
            d_i                         => s_mosi_fifo_data_in,
            we_i                        => s_mosi_wr_en,
            wr_full_o                   => s_mosi_fifo_wr_full,
            wr_empty_o                  => open,
            wr_almost_empty_o           => open,
            wr_almost_full_o            => open,
            wr_count_o                  => open,
            -- read port
            clk_rd_i                    => clk_i,
            q_o                         => s_mosi_fifo_data_out,
            rd_empty_o                  => s_mosi_fifo_rd_empty,
            rd_i                        => s_mosi_rd_en,
            rd_full_o                   => open,
            rd_almost_empty_o           => open,
            rd_almost_full_o            => open,
            rd_count_o                  => open
        );
    end generate gen_mosi_CDC;

    rmq_src_o.data <= s_mosi_fifo_data_out(31 downto 0);
    rmq_src_o.hdr <= s_mosi_fifo_data_out(34);
    rmq_src_o.last <= s_mosi_fifo_data_out(33);
    rmq_src_o.error <= s_mosi_fifo_data_out(32);
    rmq_src_o.valid <= not (s_mosi_fifo_rd_empty);

    s_mosi_rd_en <= rmq_src_i.ready and not(s_mosi_fifo_rd_empty);

end architecture rtl;
--============================================================================--
--                           End Architecture                                 --
--============================================================================--