"""

    RMQ Class

"""
import cocotb
import random

from cocotb.clock import Clock
from cocotb.triggers import RisingEdge, FallingEdge, Timer
from cocotb.result import *
from cocotb.queue import Queue, QueueFull

# RMQ Common
class RMQ:
    def __init__(self, dut, clk, if_name):
        self.dut = dut
        self.clk = clk
        self.rmq_i = getattr(dut, 'rmq_'+if_name+'_i')
        self.rmq_o = getattr(dut, 'rmq_'+if_name+'_o')
        self.fifo = Queue()

    def get_fifo_size(self):
        return self.fifo.qsize()

    def check_fifo_empty(self):
        return self.fifo.empty()

    def push_fifo(self, data):
        self.fifo.put_nowait(data)

    def pop_fifo(self):
        return self.fifo.get_nowait()

    def get_data(self):
        data = []
        for i in range(self.get_fifo_size()):
            data.append(self.pop_fifo())
        return data

"""

    RMQ Master

"""
class RMQMaster(RMQ):
    def __init__(self, dut, clk, if_name, header_size=1):
        RMQ.__init__(self, dut=dut, clk=clk, if_name=if_name)
        self.dut._log.info("Init RMQ Master")
        self.rmq_i.data.setimmediatevalue(0)
        self.rmq_i.hdr.setimmediatevalue(0)
        self.rmq_i.last.setimmediatevalue(0)
        self.rmq_i.error.setimmediatevalue(0)
        self.rmq_i.valid.setimmediatevalue(0)
        self.hdr_size = header_size-1

    def send_data(self, data):
        data_len = len(data)
        self.dut._log.info("Sending Data with length: {}".format(data_len))
        for x in range(data_len):
            if x <= self.hdr_size:
                self.push_fifo('HDR')
            if x == (data_len-1):
                self.push_fifo('LAST')
            self.push_fifo(data[x])
        self.dut._log.info("Data sent")

    @cocotb.coroutine
    def input_data_rmq(self):
        self.dut._log.info('Starting coroutine to write data')
        self.rmq_i.valid.value = 0
        fifo_data = 0
        while(True):
            yield RisingEdge(self.clk)
            while( (self.rmq_i.valid.value == 1) and (self.rmq_o.ready.value == 0)):
                yield RisingEdge(self.clk)
            self.rmq_i.hdr.value = 0
            self.rmq_i.last.value = 0
            self.rmq_i.error.value = 0
            while(self.check_fifo_empty()):
                self.rmq_i.valid.value = 0
                yield RisingEdge(self.clk)

            fifo_data = self.pop_fifo()
            if type(fifo_data) == str:
                if fifo_data == 'HDR':
                    self.rmq_i.hdr.value = 1
                    fifo_data = self.pop_fifo()
                if fifo_data == 'LAST':
                    self.rmq_i.last.value = 1
                    fifo_data = self.pop_fifo()
            self.rmq_i.data.value = fifo_data
            self.rmq_i.valid.value = 1

"""

    RMQ Slave

"""
class RMQSlave(RMQ):
    def __init__(self, dut, clk, if_name, header_size=2):
        RMQ.__init__(self, dut=dut, clk=clk, if_name=if_name)
        self.dut._log.info("Init RMQ Slave")
        self.rmq_i.ready.setimmediatevalue(0)
        self.rmq_i.pkt_ready.setimmediatevalue(0)
        self.rmq_request_c = 0
        self.rmq_word_c = 0
        self.hdr_size = header_size -1

    @cocotb.coroutine
    def output_data_rmq(self):
        self.dut._log.info('Starting coroutine to receive data')
        yield RisingEdge(self.clk)

        while(True):
            if ((self.rmq_o.valid.value == 1) and (self.rmq_i.ready.value == 1)):
                if (self.rmq_word_c <= self.hdr_size) and self.rmq_o.hdr.value == 0:
                    self.dut._log.error("Missing HEADER in reception")
                    raise SimFailure()

                self.push_fifo(self.rmq_o.data.value.integer)

                if self.rmq_o.last.value == 1:
                    self.rmq_word_c += 1
                    self.rmq_request_c = self.rmq_request_c + 1
                    self.dut._log.info("Received frame {} with length {}".format(self.rmq_request_c, self.rmq_word_c))
                    self.rmq_word_c = 0
                else:
                    self.rmq_word_c += 1
            yield RisingEdge(self.clk)

    @cocotb.coroutine
    def tready_behaviour(self, max_cycles_off=5, randomize=False):
        self.rmq_i.pkt_ready.value = 1

        if max_cycles_off == 0:

            while(True):
                self.rmq_i.ready.value = 1
                yield RisingEdge(self.clk)

        else:

            self.dut._log.info('Starting coroutine to control RMQ Ready')
            if randomize:
                max_cycles = random.randint(1, max_cycles_off)
            else:
                max_cycles = max_cycles_off

            self.rmq_i.ready.value = 0
            for x in range(max_cycles_off):
                yield RisingEdge(self.clk)

            aux_rmq_valid = 1

            while(True):
                for x in range(max_cycles_off):
                    self.rmq_i.ready.value = aux_rmq_valid
                    yield RisingEdge(self.clk)
                if aux_rmq_valid == 0:
                    aux_rmq_valid = 1
                else:
                    aux_rmq_valid = 0

                if randomize:
                    max_cycles = random.randint(1, max_cycles_off)
                else:
                    max_cycles = max_cycles_off