"""

    Test SPI RMQ BRIDGE

"""
import cocotb
import numpy as np
import time
import random
import math

from cocotb.clock import Clock
from cocotb.triggers import RisingEdge, FallingEdge, Timer
from cocotb.regression import TestFactory
from cocotb.result import *

from spi import SpiClock, SpiMaster
from rmq import RMQMaster, RMQSlave

# ------------------------------------------------------------------------------
# Some functions to support the test
# Define SPI mode
def define_spi_mode(dut, g_CPOL=0, g_CPHA=0):
    CPOL = False
    CPHA = False

    if g_CPOL == 1:
        CPOL = True
    if g_CPHA == 1:
        CPHA = True

    dut._log.info("     +-------------+")
    dut._log.info("     | CPOL | CPHA |")
    dut._log.info("     +-------------+")
    dut._log.info("MODE |  {}   |  {}   |".format(g_CPOL,g_CPHA))
    dut._log.info("     +-------------+")
    return CPOL, CPHA

# ------------------------------------------------------------------------------
# fill data
def GetSpiFrame(ctrl, rmq_id, data_len, data_width, randomize):
    spi_header = (ctrl << 24) | (rmq_id << 16) | (data_len << 8) | 0xFF
    data = []
    mask = 2**data_width - 1
    data.append(spi_header & mask)
    for i in range(data_len):
        if randomize:
            data.append(random.randint(0, (2**data_width)-1) & mask)
        else:
            if i % 8 == 0:
                data.append(0xDEADC0DE & mask)
            elif i % 8 == 1:
                data.append(0xCAFED00D & mask)
            elif i % 8 == 2:
                data.append(0xDEADCAFE & mask)
            elif i % 8 == 3:
                data.append(0xBADC0FFE & mask)
            elif i % 8 == 4:
                data.append(0xBAADF00D & mask)
            elif i % 8 == 5:
                data.append(0xDEFEC8ED & mask)
            elif i % 8 == 6:
                data.append(0xDEADBEEF & mask)
            elif i % 8 == 7:
                data.append(0xFEEDBABE & mask)
    return data

def GetSpiDummyFrame(ctrl, rmq_id, data_len, data_width):
    data = []
    spi_header = (ctrl << 24) | (rmq_id << 16) | (data_len << 8) | 0xFF
    mask = 2**data_width - 1
    data.append(spi_header & mask)
    for i in range(data_len):
        data.append(0)
    return data

# ------------------------------------------------------------------------------
# Check if data written in RMQ is ok
def check_data_written(dut, rmq_src_msg, spi_frame):
    rmq_encapsulated_msg = spi_frame[1:]
    rmq_src_msg_sized = []
    if len(rmq_encapsulated_msg) != len(rmq_src_msg)/2:
        dut._log.error("Expected length {} --- Read length {}".format(len(data_written),len(rmq_src_msg)))
        dut._log.error("Write data failed, length is wrong")
        raise SimFailure()
    else:
        dut._log.info("RMQ Data length is consistent")
        for i in range(len(rmq_encapsulated_msg)):
            rmq_src_msg_sized.append((rmq_src_msg[2*i] << 16) + rmq_src_msg[(2*i)+1])
        if (np.array_equal(rmq_src_msg_sized, rmq_encapsulated_msg)):
            dut._log.info("RMQ Data integrity is correct")
        else:
            dut._log.warning("Data RD     : {}".format(rmq_src_msg_sized))
            dut._log.warning("Data Written: {}".format(rmq_encapsulated_msg))
            dut._log.error("Write data failed, data is not the same in the SLAVE")
            raise SimFailure()

# ------------------------------------------------------------------------------
# Generate Dummy Reply
def GetRMQDummyReply(rmq_src_msg):
    rmq_snk_msg = []
    rmq_snk_msg.append((rmq_src_msg[0] << 16) + rmq_src_msg[1])
    for i in range(len(rmq_src_msg)-2):
        rmq_snk_msg.append(rmq_src_msg[i+2])
    return rmq_snk_msg

# ------------------------------------------------------------------------------
# Setup clock signal
def setup_clock(signal_name, period, units="ns"):
    cocotb.start_soon(Clock(signal_name, period, units=units).start())

# ------------------------------------------------------------------------------
# Main Test
@cocotb.coroutine
async def test(dut, n_rst_cycles, data_len, sys_freq, spi_sample_freq, spi_freq, randomize):

    # Get data to send
    data_width = dut.g_data_width.value
    g_cpol = dut.g_cpol.value
    g_cpha = dut.g_cpha.value
    g_CDC = dut.g_cdc_enable.value

    CPOL, CPHA = define_spi_mode(dut=dut, g_CPOL=g_cpol, g_CPHA=g_cpha)

    spi_frame = GetSpiFrame(ctrl=0x0, rmq_id=0x0, data_len=data_len, data_width=data_width, randomize=randomize)

    spi_master = SpiMaster(
                            dut=dut,
                            sclk=dut.spi_clk_i,
                            mosi=dut.spi_mosi_i,
                            miso=dut.spi_miso_o,
                            cs=dut.spi_cs_n_i,
                            SpiFreqz = spi_freq,
                            CPOL = CPOL,
                            CPHA = CPHA,
                            word_width=data_width
                        )

    m_rmq = RMQMaster(dut=dut, clk=dut.clk_i, if_name='snk')
    s_rmq = RMQSlave(dut=dut, clk=dut.clk_i, if_name='src')

    if g_CDC == 1:
        dut._log.warning("CDC between SPI and System")
        setup_clock(signal_name=dut.clk_i, period=(1e9/sys_freq))
        setup_clock(signal_name=dut.spi_sample_clk_i, period=(1e9/spi_sample_freq))
    else:
        dut._log.warning("NO CDC between SPI and System")
        setup_clock(signal_name=dut.clk_i, period=(1e9/sys_freq))
        setup_clock(signal_name=dut.spi_sample_clk_i, period=(1e9/sys_freq))

    cocotb.start_soon(reset(dut, n_rst_cycles))
    await delay(dut.clk_i, 1)

    while dut.rst_n_i.value != 1:
        await RisingEdge(dut.clk_i)

    cocotb.start_soon(m_rmq.input_data_rmq())
    cocotb.start_soon(s_rmq.output_data_rmq())
    cocotb.start_soon(s_rmq.tready_behaviour(max_cycles_off=0))

    await spi_master.write(spi_frame)
    await spi_master.read() #Empty fifo, don't care about data

    while not (s_rmq.rmq_frame_c == 1):
        await RisingEdge(dut.clk_i)

    rmq_src_msg = s_rmq.get_data()
    check_data_written(dut=dut, rmq_src_msg=rmq_src_msg, spi_frame=spi_frame)

    rmq_snk_msg = GetRMQDummyReply(rmq_src_msg=rmq_src_msg)
    m_rmq.send_data(rmq_snk_msg)

    spi_dummy = GetSpiDummyFrame(ctrl=0x1, rmq_id=0x0, data_len=data_len, data_width=data_width)
    await spi_master.write(spi_dummy)
    data_read = await spi_master.read()

    if len(data_read[1:]) != len(spi_frame[1:]):
        dut._log.error("Expected length {} --- Read length {}".format(len(data_read[1:]),len(spi_frame[1:])))
        dut._log.error("Read RMQ message failed")
        raise SimFailure()
    else:
        dut._log.info("RMQ message length is consistent")
        if (np.array_equal(data_read[1:], spi_frame[1:])):
            dut._log.info("RMQ message integrity is correct")
        else:
            for i in range(len(data_read)-1):
                dut._log.warning("Data TX SPI: {} ----- Data RX SPI: {}".format(hex(spi_frame[i+1]),hex(data_read[i+1])))
            dut._log.error("Incossitence between transmition and reception")
            raise SimFailure()

    await delay(dut.clk_i, 100)
    dut._log.info("Testbench finished successfully")

# ------------------------------------------------------------------------------
# Factory
factory = TestFactory(test)
factory.add_option("n_rst_cycles", [11])
factory.add_option("data_len", [4,16,32,255])
factory.add_option("sys_freq", [62.5e6])
factory.add_option("spi_sample_freq", [200e6])
factory.add_option("spi_freq", [20e6])
factory.add_option("randomize", [False, True])
factory.generate_tests()

# ------------------------------------------------------------------------------
# Reset
@cocotb.coroutine
def reset(dut, n_rst_cycles):
    dut._log.info("Asserting Sytem Reset")
    dut.rst_n_i.setimmediatevalue(0)
    dut.rmq_src_config_i.adr.setimmediatevalue(0)
    dut.rmq_src_config_i.dat.setimmediatevalue(0)
    dut.rmq_src_config_i.we.setimmediatevalue(0)
    dut.rmq_snk_config_i.adr.setimmediatevalue(0)
    dut.rmq_snk_config_i.dat.setimmediatevalue(0)
    dut.rmq_snk_config_i.we.setimmediatevalue(0)
    for i in range(n_rst_cycles):
        yield RisingEdge(dut.clk_i)
    dut.rst_n_i.value = 1
    yield RisingEdge(dut.clk_i)
    dut._log.info("Deasserting Sytem Reset")

# ------------------------------------------------------------------------------
# delay
@cocotb.coroutine
def delay(signal_name, n_cycles):
    for i in range(n_cycles):
        yield RisingEdge(signal_name)
