`timescale 1ns/1ps

`include "vme64x_bfm.svh"
`include "svec_vme_buffers.svh"

`include "vhd_wishbone_master.svh"
`include "logger.svh"
`include "mock_turtle_driver.svh"
`include "masterfip_driver.svh"
`include "fip_hardware_model.svh"
`include "fip_phy_model.svh"

`define VME_OFFSET      'h8000_0000
`define MT_BASE         `VME_OFFSET + 'h0004_0000



module main;

    parameter g_fast_load_cpu_firmware = 0;
   
    reg rst_n = 0;
    reg clk_sys = 0;
    reg clk_test = 0;
    reg spi_clk;
    reg rst;
    reg ertec_rst = 0;

    logic spi_cs_n = 1'b1;
    logic spi_mosi = 1'b0;
    logic spi_miso = 1'b0;
    logic [31:0] mosi_data;
    logic [31:0] miso_data;

    logic [4:0]  posedge_cnt = 'h0;

    logic tmp_spi_clk = 1'b0;

    logic [31:0] q_mosi_data[$];
    logic [31:0] q_miso_data[$];

    always #8ns clk_sys <= ~clk_sys;
    always #100ns tmp_spi_clk <= ~tmp_spi_clk; // 500ns for 1MHz, 100ns for 5MHz, 50ns for 10MHz, 25ns for 20MHz

    assign spi_clk = ~spi_cs_n ? tmp_spi_clk : 0;

    initial begin
        repeat(20) @(posedge clk_sys);
        rst_n = 1;
        ertec_rst = 1;
    end
 
    IVME64X VME(rst_n);

    `DECLARE_VME_BUFFERS(VME.slave);

    bit [4:0] slot_id = 8;

    // Board and FIP bus instantiation
    pulldown(fip_bus.p);
    pulldown(fip_bus.n);

    FIPBus fip_bus();
    FieldriveFPGA fieldrive();

    FipPHY U_Agent1PHY ( fip_bus.p, fip_bus.n );
    FipPHY U_Agent2PHY ( fip_bus.p, fip_bus.n );   
    FipPHY U_Agent3PHY ( fip_bus.p, fip_bus.n );
    FipPHY U_MonitorPHY ( fip_bus.p, fip_bus.n );
    FipPHY U_VirtualMasterPHY ( fip_bus.p, fip_bus.n );

    MasterfipMezzanine U_Mezzanine ( fieldrive, fip_bus );



    // the Device Under Test   
    svec_masterfip_mt_urv
        #(
          .g_simulation(1)
        )
    DUT (
      // Clock, Reset
      .rst_n_i             (rst_n),
      .clk_125m_pllref_p_i (clk_sys),
      .clk_125m_pllref_n_i (~clk_sys),
      // Defining the SPEED
      .speed_b0_i          (1'b0), // "10" -> 2.5MHz
      .speed_b1_i          (1'b1),
      // ERTEC SPI Interface
      .ertec_spi_clk_i     (spi_clk),
      .ertec_spi_mosi_i    (spi_mosi),
      .ertec_spi_cs_n_i    (spi_cs_n),
      .ertec_spi_miso_o    (spi_miso),
      .ertec_rst_i         (ertec_rst),
      // Used to connect NanoFIP
      .fd_rxcdn_i          (fieldrive.fd_rxcdn),
      .fd_rxd_i            (fieldrive.fd_rxd),
      .fd_txer_i           (fieldrive.fd_txer),
      .fd_wdgn_i           (fieldrive.fd_wdgn),
      .fd_rstn_o           (fieldrive.fd_rstn),
      .fd_txck_o           (fieldrive.fd_txck),
      .fd_txd_o            (fieldrive.fd_txd),
      .fd_txena_o          (fieldrive.fd_txena),  
      // VME Interface
      .vme_sysreset_n_i    (VME_RST_n),
      .vme_as_n_i          (VME_AS_n),
      .vme_write_n_i       (VME_WRITE_n),
      .vme_am_i            (VME_AM),
      .vme_ds_n_i          (VME_DS_n),
      .vme_gap_i           (^slot_id),
      .vme_ga_i            (~slot_id),
      .vme_berr_o          (VME_BERR),
      .vme_dtack_n_o       (VME_DTACK_n),
      .vme_retry_n_o       (VME_RETRY_n),
      .vme_retry_oe_o      (VME_RETRY_OE),
      .vme_lword_n_b       (VME_LWORD_n),
      .vme_addr_b          (VME_ADDR),
      .vme_data_b          (VME_DATA),
      .vme_irq_o           (VME_IRQ_n),
      .vme_iack_n_i        (VME_IACK_n),
      .vme_iackin_n_i      (VME_IACKIN_n),
      .vme_iackout_n_o     (VME_IACKOUT_n),
      .vme_dtack_oe_o      (VME_DTACK_OE),
      .vme_data_dir_o      (VME_DATA_DIR),
      .vme_data_oe_n_o     (VME_DATA_OE_N),
      .vme_addr_dir_o      (VME_ADDR_DIR),
      .vme_addr_oe_n_o     (VME_ADDR_OE_N)
    );

    IMockTurtleIRQ IrqMonitor (`MT_ATTACH_IRQ(DUT.cmp_mock_turtle_urv));
      IVHDWishboneMaster U_WBMaster
      (
          .clk_i(DUT.clk_62m5_sys),
          .rst_n_i(DUT.rst_n_sys)
      );

    task automatic checkSDBSignature( CBusAccessor acc );
        automatic uint64_t sig;
        automatic Logger l = Logger::get();

        const uint32_t SDB_ID = 'h5344422d;
        acc.read(0, sig);
        l.startTest("Check Wishbone communication");
        l.msg(0, $sformatf("SDB Signature: 0x%x", sig));
        if(sig != SDB_ID)
            l.fail($sformatf("Incorrect SDB signature (got 0x%x, expected 0x%x)", sig, SDB_ID));
        else
            l.pass();   
    endtask // check_sdb_signature

    task automatic checkCPUCommunication(MasterfipDriver drv);
        automatic Logger l = Logger::get();
        automatic MQueueMsg request = new (0, 5, '{}, '{});
        automatic int status;

        automatic mstrfip_hw_speed_trtlmsg response = new ();
        automatic MQueueMsg tmp = response;
        response.core = 0;
        response.slot = 5;
        request.header.flags  = `TRTL_HMQ_HEADER_FLAG_RPC | `TRTL_HMQ_HEADER_FLAG_SYNC;
        request.header.msg_id = MSTRFIP_CMD_GET_BITRATE;
        request.header.len = 2;
        request.data[0] = MSTRFIP_CMD_GET_BITRATE;

        $display("request [1]",request.tostring());
        $display("response [1]",response.tostring());

        drv.sendAndReceiveSync(request, tmp, 100, status );

        $display("tmp ",tmp.tostring());

        // TODO: The right one, for bitrate is the response.header.msg_id, because
        // this information is in the header part of the message and not in
        // payload. This is to be fixed.
        if(status < 0)
        begin
        	l.fail("No response from CPU 0 within 100us");
        	return;
        end else begin

          if ( tmp.data[0] != MSTRFIP_REP_BITRATE )
          begin
            l.fail($sformatf("  Incorrect response from the RT, expected 0x10, got 0x%x", tmp.data[0] ) );
            return;
          end
//      response.data_unpack();
//      l.msg(0, $sformatf("Got bitrate from CPU0: %d", response.data[0]));
        end
        l.pass();
    endtask

    always @(posedge spi_clk) begin
        miso_data <= {miso_data[30:0], spi_miso};
    end

    task automatic loadCPUFirmwares(MasterfipDriver drv);
        automatic Logger l = Logger::get();
        const string cpu0_fw = "masterfip-rt-ba.bin"; //"masterfip-rt-ba-sim.bin";
        const string cpu1_fw = "masterfip-rt-cmd.bin";
        l.startTest("Load CPU Firmwares & check CPU communication");
        drv.m_mt.rom.pretty_print();
      
        if (!g_fast_load_cpu_firmware)
        begin
        	  l.msg(0, $sformatf("Loading '%s' to CPU0", cpu0_fw) );
        	  drv.m_mt.load_firmware(0, cpu0_fw);
        	  l.msg(0, $sformatf("Loading '%s' to CPU1", cpu1_fw) );
            drv.m_mt.load_firmware(1, cpu1_fw);
        end
        l.msg(0, "Resetting CPUs");
        drv.m_mt.reset_core(0, 0);
        drv.m_mt.reset_core(1, 0);
        #100us; // give some time for the firmware to boot up
        checkCPUCommunication(drv);
    endtask // loadCPUFirmwares

    CBusAccessor_VME64x acc;

    // counter that counts the posedge of spi_clk
    always @(posedge spi_clk)
    begin
        if (spi_cs_n)
            posedge_cnt <= 0;
        else
            posedge_cnt <= posedge_cnt + 1;
    end

    /*
    * Tasks for SPI
    */

    logic [7:0]  operation;
    logic [7:0]  data_len;
    logic [7:0]  rmq_id;
    logic [31:0] basic_info;

    // Write SPI mosi data task. Used to generate MOSI data
    task write_spi(input logic [31:0] mosi_spi_data, input int data_length);
        int i;
        for (i=data_length;i>0;i--)
        begin
            @(posedge spi_clk);
            spi_mosi = mosi_spi_data[i-1];
        end
    endtask

    // Generate the frame that will be written in the RMQ
    task write_to_rmq();
        int j;
        operation  = 8'h00;
        rmq_id     = $urandom_range(0,6);
        data_len   = $urandom_range(1,16);
        basic_info = {operation, rmq_id, data_len, 8'h00}; // < wr/rd | rmq_id | len | unused >

        // First send the Basic Info
        write_spi(basic_info, 32);

        // Then send the 3 header words and payload
        for (j=3+data_len;j>0;j--) 
        begin
            mosi_data = $urandom_range(1,2048);
            write_spi(mosi_data, 32);
        end
        @(posedge spi_clk);
        spi_mosi = 'b0;
    endtask // write_to_rmq

    task write_shorter_to_rmq();
        int j;
        operation  = 8'h00;
        rmq_id     = $urandom_range(0,6);
        data_len   = $urandom_range(1,16);
        basic_info = {operation, rmq_id, data_len, 8'h00}; // < wr/rd | rmq_id | len | unused >

        // First send the Basic Info
        write_spi(basic_info, 32);

        // Then send the 3 header words and payload
        for (j=3+data_len;j>1;j--) 
        begin
            mosi_data = $urandom_range(1,2048);
            write_spi(mosi_data, 32);
        end
        @(posedge spi_clk);
        spi_mosi = 'b0;
    endtask

    task write_longer_to_rmq();
        int j;
        operation  = 8'h00;
        rmq_id     = $urandom_range(0,6);
        data_len   = $urandom_range(1,16);
        basic_info = {operation, rmq_id, data_len, 8'h00}; // < wr/rd | rmq_id | len | unused >

        // First send the Basic Info
        write_spi(basic_info, 32);

        // Then send the 3 header words and payload
        for (j=3+data_len+5;j>0;j--) 
        begin
            mosi_data = $urandom_range(1,2048);
            write_spi(mosi_data, 32);
        end
        @(posedge spi_clk);
        spi_mosi = 'b0;
    endtask

    task read_spi(input logic [31:0] basic_info);
        write_spi(basic_info, 32);
        @(posedge DUT.cmp_mt_profip_translator.s_rd_state == 9); // END_OF_FRAME
        @(negedge DUT.cmp_mt_profip_translator.s_ready);
        @(posedge spi_clk); // just for display reasons
    endtask // read_spi

    task read_from_rmq();
        operation  = 8'h01;
        basic_info = {operation, rmq_id, 8'h00, 8'h00};
        
        write_spi(basic_info, 32);
        // Then we shifting MISO data, 32 spi clock cycles after the end of the frame
        @(posedge DUT.cmp_mt_profip_translator.s_rd_state == 9); // END_OF_FRAME
        @(negedge DUT.cmp_mt_profip_translator.s_ready);
        @(posedge spi_clk); // just for display reasons
    endtask

    task read_longer_from_rmq();
        operation  = 8'h01;
        basic_info = {operation, rmq_id, 8'h00, 8'h00};
        
        write_spi(basic_info, 32);
        // Then we shifting MISO data, 32 spi clock cycles after the end of the frame
        @(posedge DUT.cmp_mt_profip_translator.s_rd_state == 9);
        @(negedge DUT.cmp_mt_profip_translator.s_ready);
        repeat(99) @(posedge spi_clk);
    endtask

    task read_shorter_from_rmq();
        operation  = 8'h01;
        basic_info = {operation, rmq_id, 8'h00, 8'h00};
        
        write_spi(basic_info, 32);
        // Then we shifting MISO data, 32 spi clock cycles after the end of the frame
        @(posedge DUT.cmp_mt_profip_translator.s_rd_state == 6);
    endtask


    int mosi_cnt = 0; // index to the mosi data queue
    int miso_cnt = 0; // index to the miso data queue

    logic [15:0] shreg_msb, shreg_lsb;
    logic [31:0] shreg = 32'h0;

    // Collects the data that is written in RMQ. From each data, the last 16-bits are valid
    always @(posedge clk_sys)
    begin
        if (DUT.rmq_src_out.valid & !DUT.rmq_src_out.error & DUT.rmq_src_in.ready)
        begin
            shreg_msb <= DUT.rmq_src_out.data[15:0];
            shreg_lsb <= shreg_msb;
            shreg     <= {shreg_lsb, shreg_msb}; 
        end
    end

    // Push back in the miso queue the miso data
    always @(posedge DUT.cmp_mt_profip_translator.s_ready)
    begin
        q_miso_data.push_back(miso_data);
    end

    // Push back in the mosi queue the MOSI data
    always @(posedge DUT.cmp_mt_profip_translator.s_wr_data_valid)
    begin
        q_mosi_data.push_back(DUT.cmp_mt_profip_translator.s_data_reg_o);
    end

    // -----------------------------------
    //    Assertions
    // ------------------------------------


    // -----------------------------------
    // -- WRITE SIDE OF RMQ
    // -----------------------------------

    // Rule write_1 : Ensure that the incoming mosi data is the same as
    // the one that is generated by the SPI in mt_profip_translator
    always @(posedge DUT.cmp_mt_profip_translator.s_wr_data_valid)
    begin
        if (operation == 0)
        begin
          assert (q_mosi_data[mosi_cnt] == DUT.cmp_mt_profip_translator.s_data_reg_o)
              else $error("Mismatch in MOSI data, position : %d, expected : %h, received : %h ",mosi_cnt, q_mosi_data[mosi_cnt], DUT.cmp_mt_profip_translator.s_data_reg_o);
          if (mosi_cnt != (4 + data_len))
              mosi_cnt <= mosi_cnt + 1;
          else
              mosi_cnt <= 0;
        end
    end

    // Rule write_2: Check that when valid is HIGH in RMQ, after 1 or 2 system clock cycles, the rmq data is the expected one
    property checking_mosi_rmq_data;
        @(posedge clk_sys) ($rose(DUT.rmq_src_out.valid) & !(DUT.rmq_src_out.error)) |-> ##2 (shreg == DUT.cmp_mt_profip_translator.s_mosi_data);
    endproperty

    assert property (checking_mosi_rmq_data)
        else $error(" MISMATCH in write RMQ data : shreg = %h and mosi_data = %h",shreg, DUT.cmp_mt_profip_translator.s_mosi_data);

    // Rule write_3 : Check that the Basic Info data, is the expected one, 32 spi clock cycles after CS is LOW
    property check_basic_info;
        @(posedge spi_clk) ($fell(spi_cs_n)) |-> ##[32:33] (basic_info == DUT.cmp_mt_profip_translator.s_data_reg_o);
    endproperty

    assert property (check_basic_info)
        else $error("MISMATCH in Basic Info : expected : %h, received : %h",basic_info, DUT.cmp_mt_profip_translator.s_data_reg_o);


    // -----------------------------------
    // -- READ SIDE OF RMQ
    // -----------------------------------
    
    // Rule read_1: check that the data that is sending through MISO with SPI is the expected one
    property check_miso_spi_data;
        @(posedge spi_clk) ($rose(DUT.cmp_mt_profip_translator.s_ready) |-> ##32 ((DUT.cmp_mt_profip_translator.s_spi_miso_data) == miso_data));
    endproperty

    assert property (check_miso_spi_data)
        else $error("MISMATCH in SPI MISO : expected = %h, received = %h",DUT.cmp_mt_profip_translator.s_spi_miso_data, miso_data);


    logic [31:0] send_rst[11];
    logic [31:0] send_worldfip_speed[7];
    logic [31:0] send_hw_config_set[12];
    logic [31:0] send_msg_1[];
    logic [31:0] send_msg_2[];
    logic [31:0] send_msg_3[];
    logic [31:0] send_msg_4[];
    logic [31:0] send_msg_5[];
    logic [31:0] send_msg_6[];


    initial begin
        send_rst = '{'h60700, 'h3000001, 'h10007, 'h8014c90, 'h3, 'h0, 'hff8e, 'h1921d, 'h202952ec, 'h8000d34, 'h2029535c};

        send_worldfip_speed = '{'h50300, 'h1000000, 'h20003, 'h10b5a, 'h1, 'h0, 'h2029535c};

        send_hw_config_set = '{'h50800, 'h2000000, 'h30008, 'h0, 'h2, 'h0, 'h0, 'h0, 'h1, 'hbb8, 'h2d50, 'h28};

        send_msg_1 = '{'h56400, 'ha000000, 'h40064, 'h431bde83, 'ha, 'h0, 'h125d90, 'h1, 'h55, 'h8014d42, 'h22,
                       'h8014c98, 'h201dfb30, 'h201dfcc4, 'h80000013, 'h6, 'h201dfb30, 'h8014c98, 'h12c, 'h201e79ac, 'h17562,
                        'h1fa7, 'h202952ec, 'h8000d34, 'h2029535c, 'h202952ec, 'h8014d1c, 'h0, 'h8014d60, 'h2f38, 'h202952b4,
                        'h8014d42, 'h2029535c, 'h202952ec, 'h8014d1c, 'h201de574, 'h201de504, 'h23fffe00, 'hdc000221, 'h21, 'h21,
                        'hd, 'hfffffff7, 'h8014d60, 'h20295664, 'h201def00, 'h8014d84, 'h22, 'h8014d8c, 'h202c12f0, 'h305b1b5d, 'h0,
                        'h32335b1b, 'h4b4f5b6d, 'h305b1b5d, 'h614d206d, 'h636f7263, 'h656c6379, 'h65726320, 'h64657461, 'hdead0a0d, 'hdeadbeef,
                        'h1, 'h202a8600,  'h8000ca4, 'h8000d34, 'h1, 'h202a8600, 'h8000ca4, 'h8000d34, 'h202c169c, 'h8000d34, 'h0, 'h2, 'h21,
                        'h202c137c, 'h0, 'h2, 'h1, 'h202c39a8, 'h8014e94, 'h201e25d0, 'h21, 'h201ef72c, 'h0, 'h8014db8, 'h1, 'h27, 'h0, 'h100, 'h23fffde0,
                        'h21, 'h5c64, 'h5bdc, 'h1, 'h0, 'h2029b75c, 'h2, 'h1, 'h2029b75c, 'h1765, 'h201f506c, 'h0, 'h1765};

        send_msg_2 = '{ 'h56400,  'ha000000,  'h50064, 'h0, 'ha, 'h0, 'h125d90, 'h56, 'haa, 'hfffff7ea, 'h0,  'h0, 'h0, 'h0,        
                        'h0,      'h0,        'h0,     'h0, 'h0, 'h0, 'h0,      'h0 , 'h0 , 'h0,        'h0,  'h0, 'h0, 'h0,        
                        'h0,      'h0,        'h0,     'h0, 'h0, 'h0, 'h0,      'h0,  'h0,  'h0,        'h0,  'h0, 'h0, 
                        'h0,      'h0,        'h0,     'h0, 'h0, 'h0, 'h0,      'h0,  'h0,  'h0,        'h0,  'h0, 'h0, 
                        'h0,      'h0,        'h0,     'h0, 'h0, 'h0, 'h0,      'h0,  'h0,  'h0,        'h0,  'h0, 'h0, 
                        'h0,      'h0,        'h0,     'h0, 'h0, 'h0, 'h0,      'h0,  'h0,  'h0,        'h0,  'h0, 'h0, 
                        'h0,      'h0,        'h0,     'h0, 'h0, 'h0, 'h0,      'h0,  'h0,  'h0,        'h0,  'h0, 'h0,
                        'h5bdc,   'h1,        'h0,     'h2029b75c, 'h2, 'h1, 'h2029b75c, 'h1765, 'h201f506c, 'h0, 'h1765};

        send_msg_3 = '{'h56400, 'ha000000, 'h60064, 'h0, 'ha, 'h0, 'h125d90, 'hab, 'hff, 'h0, 'h0, 'h0, 'h0, 
                        'h0,  'h0, 'h0, 'h0, 'h0, 'h0,    'h0, 'h0, 'h0,        'h0, 'h0, 'h0,       'h0, 'h0, 'h0,      
                        'h0,  'h0, 'h0, 'h0, 'h0, 'h0,    'h0, 'h0, 'h0,        'h0, 'h0, 'h0,       'h0, 'h0, 'h0,    
                        'h0,  'h0, 'h0, 'h0, 'h0, 'h0,    'h0, 'h0, 'h0,        'h0, 'h0, 'h0,       'h0, 'h0, 'h0,  
                        'h0,  'h0, 'h0, 'h0, 'h0, 'h0,    'h0, 'h0, 'h0,        'h0, 'h0, 'h0,       'h0, 'h0, 'h0,   
                        'h0,  'h0, 'h0, 'h0, 'h0, 'h0,    'h0, 'h0, 'h0,        'h0, 'h0, 'h0,       'h0, 'h0, 'h0,   
                        'h0,  'h0, 'h0, 'h0, 'h0, 'h5bdc, 'h1, 'h0, 'h2029b75c, 'h2, 'h1, 'h2029b75c, 'h1765, 'h201f506c, 'h0, 'h1765};

        send_msg_4 = '{'h50800, 'h2000000, 'h70008, 'h0, 'h2, 'h0, 'h0, 'h0,       'h1,        'hbb8    , 'h2d50    ,'h28};
        send_msg_5 = '{'h50700, 'h5000000, 'h80007, 'h0, 'h5, 'h0, 'h2, 'h1050000, 'h8100007c, 'h1060000, 'h8200007c};

        send_msg_6 = '{'h51900, 'h4000000, 'h90019, 'h0, 'h4, 'h0, 'h3d0900, 'h3, 'h1, 'h0, 'h1, 'h0, 'h0,
                       'h0,     'h0,        'h1,    'h1, 'h2, 'h0, 'h0,      'h0, 'h0, 'h4, 'h3d0900, 'h1, 'h0, 'h0, 'h0, 'h0};



    end


    // Main testbench
    initial begin
        automatic MockTurtleDriver drv;
        automatic MasterfipDriver fip; 
        automatic Logger l = Logger::get();

        acc = new(VME.tb);
        /* map func0 to 0x80000000, A32 */
        acc.write('h7ff63, 'h80, A32|CR_CSR|D08Byte3);
        acc.write('h7ff67, 0, CR_CSR|A32|D08Byte3);
        acc.write('h7ff6b, 0, CR_CSR|A32|D08Byte3);
        acc.write('h7ff6f, 36, CR_CSR|A32|D08Byte3);
        acc.write('h7ff33, 1, CR_CSR|A32|D08Byte3);
        acc.write('h7fffb, 'h10, CR_CSR|A32|D08Byte3); /* enable module (BIT_SET = 0x10) */

        acc.set_default_modifiers(A32 | D32 | SINGLE);
        acc.set_default_xfer_size(A32 | D32 | SINGLE);

        drv = new (acc, `MT_BASE, IrqMonitor);
        fip = new (drv);


        drv.init();
        drv.enable_hmqi_irq ( 0, 0, 1 );
        drv.enable_console_irq ( 0, 1 );
        drv.enable_console_irq ( 1, 1 );
        $display("Before loading the firmware!");

        loadCPUFirmwares(fip);
        $display("End of loading firmware\n");

// START

        //////////////////////////////////////////////////////

        $display(" **** Starting MSTRFIP RESET **** ");
        @(negedge tmp_spi_clk);
        spi_cs_n = 'b0;
        foreach (send_rst[i])
            write_spi(send_rst[i],32);
        @(posedge spi_clk);
        $display("FRAME WRITTEN TO RMQ : %p\n", q_mosi_data);
        spi_cs_n = 'b1;

        wait (DUT.ertec_rmq_status_o != 'h7F);


        $display("**** Start reading the 1st frame ****\n");
        @(negedge tmp_spi_clk);
        spi_cs_n = 'b0;
        read_spi('h1060000);
        spi_cs_n = 'b1;
        $display("FRAME READ FROM RMQ : %p\n", q_miso_data);
        #1us;

        q_mosi_data.delete();
        q_miso_data.delete();


        //////////////////////////////////////////////////////

        $display(" **** Starting WORLDFIP SPEED  **** ");
        @(negedge tmp_spi_clk);
        spi_cs_n = 'b0;
        foreach (send_worldfip_speed[i])
            write_spi(send_worldfip_speed[i],32);
        @(posedge spi_clk);
        $display("FRAME WRITTEN TO RMQ : %p\n", q_mosi_data);
        spi_cs_n = 'b1;

        wait (DUT.ertec_rmq_status_o != 'h7F);


        $display("**** Start reading the 2nd frame ****\n");
        @(negedge tmp_spi_clk);
        spi_cs_n = 'b0;
        read_spi('h1050000);
        spi_cs_n = 'b1;
        $display("FRAME READ FROM RMQ : %p\n", q_miso_data);
        #1us;

        q_mosi_data.delete();
        q_miso_data.delete();

        ///////////////////////////////////////////////////////

        $display(" **** Starting HW CONFIG SET  **** ");
        @(negedge tmp_spi_clk);
        spi_cs_n = 'b0;
        foreach (send_hw_config_set[i])
            write_spi(send_hw_config_set[i],32);
        @(posedge spi_clk);
        $display("FRAME WRITTEN TO RMQ : %p\n", q_mosi_data);
        spi_cs_n = 'b1;

        wait (DUT.ertec_rmq_status_o != 'h7F);


        $display("**** Start reading the 3rd frame ****\n");
        @(negedge tmp_spi_clk);
        spi_cs_n = 'b0;
        read_spi('h1050000);
        spi_cs_n = 'b1;
        $display("FRAME READ FROM RMQ : %p\n", q_miso_data);
        #1us;

        q_mosi_data.delete();
        q_miso_data.delete();

        /////////////////////////////////////////////////////////

        $display(" **** Starting 1st MESSAGE  **** ");
        @(negedge tmp_spi_clk);
        spi_cs_n = 'b0;
        foreach (send_msg_1[i])
            write_spi(send_msg_1[i],32);
        @(posedge spi_clk);
        @(posedge spi_clk); // KB: added it only for debug something
        $display("FRAME WRITTEN TO RMQ : %p\n", q_mosi_data);
        spi_cs_n = 'b1;

        wait (DUT.ertec_rmq_status_o != 'h7F);


        $display("**** Start reading the 4th frame ****\n");
        @(negedge tmp_spi_clk);
        spi_cs_n = 'b0;
        read_spi('h1050000);
        spi_cs_n = 'b1;
        $display("FRAME READ FROM RMQ : %p\n", q_miso_data);
        #1us;

        q_mosi_data.delete();
        q_miso_data.delete();
        
        /////////////////////////////////////////////////////////
/*
        $display(" **** Starting 2nd MESSAGE  **** ");
        @(negedge tmp_spi_clk);
        spi_cs_n = 'b0;
        foreach (send_msg_2[i])
            write_spi(send_msg_2[i],32);
        @(posedge spi_clk);
        $display("FRAME WRITTEN TO RMQ : %p\n", q_mosi_data);
        spi_cs_n = 'b1;

        wait (DUT.ertec_rmq_status_o != 'h7F);


        $display("**** Start reading the 5th frame ****\n");
        @(negedge tmp_spi_clk);
        spi_cs_n = 'b0;
        read_spi('h1050000);
        spi_cs_n = 'b1;
        $display("FRAME READ FROM RMQ : %p\n", q_miso_data);

        q_mosi_data.delete();
        q_miso_data.delete();

        /////////////////////////////////////////////////////////

        $display(" **** Starting 3rd MESSAGE  **** ");
        @(negedge tmp_spi_clk);
        spi_cs_n = 'b0;
        foreach (send_msg_3[i])
            write_spi(send_msg_3[i],32);
        @(posedge spi_clk);
        $display("FRAME WRITTEN TO RMQ : %p\n", q_mosi_data);
        spi_cs_n = 'b1;

        wait (DUT.ertec_rmq_status_o != 'h7F);


        $display("**** Start reading the 6th frame ****\n");
        @(negedge tmp_spi_clk);
        spi_cs_n = 'b0;
        read_spi('h1050000);
        spi_cs_n = 'b1;
        $display("FRAME READ FROM RMQ : %p\n", q_miso_data);
        #1us;

        q_mosi_data.delete();
        q_miso_data.delete();

        
        /////////////////////////////////////////////////////////

        $display(" **** Starting 4th MESSAGE  **** ");
        @(negedge tmp_spi_clk);
        spi_cs_n = 'b0;
        foreach (send_msg_4[i])
            write_spi(send_msg_4[i],32);
        @(posedge spi_clk);
        $display("FRAME WRITTEN TO RMQ : %p\n", q_mosi_data);
        spi_cs_n = 'b1;

        wait (DUT.ertec_rmq_status_o != 'h7F);


        $display("**** Start reading the 7th frame ****\n");
        @(negedge tmp_spi_clk);
        spi_cs_n = 'b0;
        read_spi('h1050000);
        spi_cs_n = 'b1;
        $display("FRAME READ FROM RMQ : %p\n", q_miso_data);

        q_mosi_data.delete();
        q_miso_data.delete();

        /////////////////////////////////////////////////////////

        $display(" **** Starting 5th MESSAGE  **** ");
        @(negedge tmp_spi_clk);
        spi_cs_n = 'b0;
        foreach (send_msg_5[i])
            write_spi(send_msg_5[i],32);
        @(posedge spi_clk);
        $display("FRAME WRITTEN TO RMQ : %p\n", q_mosi_data);
        spi_cs_n = 'b1;

        wait (DUT.ertec_rmq_status_o != 'h7F);


        $display("**** Start reading the 8th frame ****\n");
        @(negedge tmp_spi_clk);
        spi_cs_n = 'b0;
        read_spi('h1050000);
        spi_cs_n = 'b1;
        $display("FRAME READ FROM RMQ : %p\n", q_miso_data);
        #1us;

        q_mosi_data.delete();
        q_miso_data.delete();

        
        /////////////////////////////////////////////////////////

        $display(" **** Starting 6th MESSAGE  **** ");
        @(negedge tmp_spi_clk);
        spi_cs_n = 'b0;
        foreach (send_msg_6[i])
            write_spi(send_msg_6[i],32);
        @(posedge spi_clk);
        $display("FRAME WRITTEN TO RMQ : %p\n", q_mosi_data);
        spi_cs_n = 'b1;


        wait (DUT.ertec_rmq_status_o != 'h7F);


        $display("**** Start reading the 9th frame ****\n");
        @(negedge tmp_spi_clk);
        spi_cs_n = 'b0;
        read_spi('h1050000);
        spi_cs_n = 'b1;
        $display("FRAME READ FROM RMQ : %p\n", q_miso_data);

        q_mosi_data.delete();
        q_miso_data.delete();


*/
// END

/*
        for (int i=0;i<40;i++)
        begin

            $display("**** Start writing the %d frame ****\n",i+1);
            spi_cs_n = 'b0;
            write_to_rmq();
            spi_cs_n = 'b1;
            #1ms;

            $display("FRAME WRITTEN IN RMQ : %p\n", q_mosi_data);
            mosi_cnt <= 0;

            $display("**** Start reading the %d frame ****\n",i+1);
            spi_cs_n = 'b0;
            read_from_rmq();
            spi_cs_n = 'b1;
            $display("FRAME READ FROM RMQ : %p", q_miso_data);
            #1ms;

            assert (q_mosi_data[4:data_len+4] == q_miso_data[2:data_len+2])
                else $error("Mismatch between two queues");

            q_mosi_data.delete();
            q_miso_data.delete();
            mosi_cnt <= 0;

        end
*/
  
        $finish;

    end


endmodule // main

